#include <cleanup.h>
#include <loader.h>

p_cleanup_list cl_list=NULL;
mlock_t cl_lock;

int cleanup_init(void)
{
	MUTEX_CREATE(cl_lock, "Cleanup");
	return 1;
}

/* Add a cleanup handler */
int cleanup_add_handler(proc_cleanup handler, void *priv)
{
	p_cleanup_list n=NULL;

	if ( !handler ) return 0;

	if ( !(n=malloc(sizeof(cleanup_list_t))) ){
		return 0;
	}

	n->cleanup=handler;
	n->priv=priv;
	
	CRITICAL_START(cl_lock);
	cl_list=list_append(n, cl_list);
	CRITICAL_STOP(cl_lock);
	return 1;
}

/* Call exit, but call all cleanup handlers first */
void cleanup(int code)
{
	p_cleanup_list tmp=NULL;
	p_cleanup_list foo=NULL;
	p_cleanup_list last=NULL;

	print_raw( CRLF "CLEANING UP" CRLF );
	
	/* We need to cleanup backwards */
	for (tmp=cl_list; tmp; tmp=(p_cleanup_list)tmp->l.next) {
		last=tmp;
	}

	CRITICAL_START(cl_lock);
	for (tmp=last; tmp; tmp=(p_cleanup_list)tmp->l.prev) {
		tmp->cleanup(code, tmp->priv);
	}


	/* Special case: plugin unloading.
	* Unloading plugins makes some cleanup handlers
	* disappear. We gotta unload 'em last of all */
	loader_cleanup(code, NULL);
	
	/* Cleanup the cleanup handlers */
	for (tmp=cl_list; tmp;) {
		foo=tmp;
		tmp=(p_cleanup_list)tmp->l.next;
		free(foo);
		foo=NULL;
	}
	CRITICAL_STOP(cl_lock);
	MUTEX_DESTROY(cl_lock);
	
	print_out("Exiting with code %i" CRLF, code);
	exit(code);
}


/* Generic cleanup handler for locks */
void cleanup_lock(int code, void *priv)
{
	pmlock x=priv;
	if ( priv == NULL ) return;

	//print_out("Cleaning up lock at address 0x%x" CRLF, priv);

	MUTEX_PDISABLE(x);
	MUTEX_PDESTROY(x);
	return;
}

/* Generic cleanup handler for list items with ONE pointer */
void cleanup_arg(int code, void *priv)
{
	p_config_arg foo,bar;

	for (foo=(p_config_arg)priv; foo;) {
		bar=foo;
		foo=(p_config_arg)bar->l.next;
		if (bar->arg) free(bar->arg);
		free(bar);
		bar=NULL;
	}
}

/* Generic cleanup handler for list items with TWO pointers */
void cleanup_kv(int code, void *priv)
{
	p_kv foo,bar;

	for (foo=(p_kv)priv; foo;) {
		bar=foo;
		foo=(p_kv)bar->l.next;
		if (bar->key) free(bar->key);
		if (bar->val) free(bar->val);
		free(bar);
		bar=NULL;
	}
}


/* Generic cleanup handler for list items with TWO pointers */
void cleanup_c_capdev(int code, void *priv)
{
	p_c_capdev foo,bar;

	for (foo=(p_c_capdev)priv; foo;) {
		bar=foo;
		foo=(p_c_capdev)bar->l.next;
		if (bar->type) free(bar->type);
		if (bar->source) free(bar->source);
		if (bar->filter) free(bar->filter);
		free(bar);
		bar=NULL;
	}
}
